// a ScreenSaver module by Stephan Assmus, Ingo Weinhold

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include <Alert.h>
#include <Application.h>
#include <CheckBox.h>
#include <Directory.h>
#include <Entry.h>
#include <FindDirectory.h>
#include <image.h>
#include <MenuBar.h>
#include <MenuField.h>
#include <MenuItem.h>
#include <Path.h>
#include <PopUpMenu.h>
#include <TabView.h>
#include <Screen.h>
#include <Slider.h>
#include <StringView.h>
#include <TextView.h>
#include <View.h>
#include <Window.h>
#include <WindowScreen.h>

#include "BrainWash.h"
#include "Dimensions.h"
#include "Functions.h"
#include "Threads.h"
#include "ViewDisplay.h"
#include "WindowScreenDisplay.h"

extern "C" { 
_EXPORT BScreenSaver *instantiate_screen_saver(BMessage *msg, image_id id); 
}

// version
const char *cmd_version = "BrainWash 1.4.0";

// config message field names
const char *cycle_speed_name = "CYCLE_SPEED";
const char *drawing_gap_name = "DRAWING_GAP";
const char *drawing_speed_name = "DRAWING_SPEED";
const char *screen_resolution_name = "SCREEN_RESOLUTION";
const char *compatibility_mode_name = "COMPATIBILITY_MODE";

// message what field values
enum {
	MSG_COMPATIBILITY_MODE_CHANGED	= 'cmch',
};

// min/max/default values for settings
enum
{
	MIN_CYCLE_SPEED			= 1,
	MAX_CYCLE_SPEED			= 10,
	DEF_CYCLE_SPEED			= 5,

	MIN_DRAWING_GAP			= 0,
	MAX_DRAWING_GAP			= 10,
	DEF_DRAWING_GAP			= 5,

	MIN_DRAWING_SPEED		= 1,
	MAX_DRAWING_SPEED		= 9,
	DEF_DRAWING_SPEED		= 5,

	DEF_SCREEN_RESOLUTION	= B_8_BIT_640x480,
};

// a pointer to the global settings and the default settings
Settings *settings;
Settings defaults = { DEF_CYCLE_SPEED, DEF_DRAWING_GAP, DEF_DRAWING_SPEED,
					  DEF_SCREEN_RESOLUTION, false };


// helpful functions

// add a tab page to a tab view
inline BView *add_tab_page(BTabView *tv_settings, BRect frame,
						   const char *label)
{
	BTab *tab = new BTab();
	BView *view = new BView(frame, 0, B_FOLLOW_ALL, B_WILL_DRAW);
	view->SetViewColor(tv_settings->ViewColor());
	tv_settings->AddTab(view, tab);
	tab->SetLabel(label);
	return view;
}


// classes

// SimpleSlider

class SimpleSlider : public BSlider
{
public:
	SimpleSlider(BRect frame, const char *label, int min, int max, int def,
				 const char *limlabl, const char *limlabr,
				 int32 *targetValue = 0)
		: BSlider(frame, 0, label, 0, min, max),
		  fTargetValue(targetValue)
	{
		ResizeToPreferred();
		SetLimitLabels(limlabl, limlabr);
		SetHashMarkCount(max - min + 1);
		SetHashMarks(B_HASH_MARKS_BOTTOM);
		BSlider::SetValue(def);
	}

	virtual void SetValue(int32 value)
	{
		if (fTargetValue)
			*fTargetValue = value;
		BSlider::SetValue(value);
	}
private:
	int32 *fTargetValue;
};


// SimpleCheckBox

class SimpleCheckBox : public BCheckBox
{
public:
	SimpleCheckBox(BRect frame, const char *label, bool value,
				   bool *targetValue = NULL, BMessage *message = NULL)
		: BCheckBox(frame, NULL, label, message),
		  fTargetValue(targetValue)
	{
		ResizeToPreferred();
		if (value)
			SetValue(B_CONTROL_ON);
		else
			SetValue(B_CONTROL_OFF);
	}

	virtual void SetValue(int32 value)
	{
		if (fTargetValue)
			*fTargetValue = (value == B_CONTROL_ON);
		BCheckBox::SetValue(value);
	}

private:
	bool*	fTargetValue;
};


// ScreenResMenuField
// a simple class that lets the user set the screen resolution
class ScreenResMenuField : public BMenuField
{
public:
	ScreenResMenuField(BRect frame, const char *label, uint32 res,
					   uint32 *target_value = 0)
		: BMenuField(frame, NULL, label, new BPopUpMenu("resolution")) 
	{
		this->target_value = target_value;
		BMenu *menu = Menu();
		BMenuItem *item;
		// the first item - current workspace dimension
		BMessage *msg = new BMessage(VALUE_CHANGED);
		msg->AddInt32("value", CURRENT_WORKSPACE_DIMENSIONS);
		item = new BMenuItem("workspace res.", msg);
		menu->AddItem(item);

		int32 mark = 0;
		// add an item for each supported resolution
		for (int32 i = 0; i < (int32)color_spaces_count; i++)
		{
			status_t error;
			BWindowScreen *screen = new BWindowScreen("test",
				color_spaces[i], &error);
			if (error != B_OK)
			{
				// not supported
				delete screen;
				screen = 0;
			}
			else
			{
				// supported - add a menu item
				delete screen;
				screen = 0;
				msg = new BMessage(VALUE_CHANGED);
				msg->AddInt32("value", color_spaces[i]);
				item = new BMenuItem(color_space_names[i], msg);
				menu->AddItem(item);
				if (res == color_spaces[i])
					mark = menu->IndexOf(item);
			}
		}
		// mark the current resolution
		if ((item = menu->ItemAt(mark)))
		{
			item->SetMarked(true);
			item->Message()->FindInt32("value", (int32 *)&resolution);
		}
	};
	~ScreenResMenuField() {};

	virtual void MessageReceived(BMessage *message)
	{
		switch (message->what)
		{
			case VALUE_CHANGED:
				message->FindInt32("value", (int32 *)&resolution);
				*target_value = resolution;
				break;
			default:
				BMenuField::MessageReceived(message);
				break;
		}
	};
	
	virtual void AttachedToWindow()
	{
		BMenu *menu = Menu();
		for (int32 i = 0; BMenuItem *item = menu->ItemAt(i); i++)
			item->SetTarget(this);
		BMenuField::AttachedToWindow();
	};

	void SetResolution(uint32 resolution)
	{
		// find the menu item for that resolution
		BMenu *menu = Menu();
		for (int32 i = 0; BMenuItem* item = menu->ItemAt(i); i++) {
			BMessage* message = item->Message();
			uint32 value;
			if (message
				&& message->FindInt32("value", (int32*)&value) == B_OK
				&& value == resolution) {
				// found the menu item
				item->SetMarked(true);
				this->resolution = resolution;
				return;
			}
		}
	}

private:
	uint32 *target_value;
	uint32	resolution;

	// message what value - used if a value has changed
	enum { VALUE_CHANGED = 'vach' };
};


// BrainLooper

class BrainLooper : public BLooper {
public:
	BrainLooper(BrainWashSaver* saver)
		: fSaver(saver)
	{
	}

	virtual void MessageReceived(BMessage* message)
	{
		if (message->what == MSG_COMPATIBILITY_MODE_CHANGED)
			fSaver->CompatibilityModeChanged();
	}

private:
	BrainWashSaver*	fSaver;
};


// BrainWashSaver - our main class

// constructor
BrainWashSaver::BrainWashSaver(BMessage *msg, image_id id)
	: BScreenSaver(msg, id),
	  fLooper(NULL),
	  fScreenResolutionMenu(NULL),
	  fCompatibilityModeBox(NULL),
	  fDisplay(NULL),
	  fCycleThread(NULL),
	  fDrawThread(NULL)
{
	// set settings defaults
	settings = new Settings;
	*settings = defaults;
	// get settings values from message
	if (msg->HasInt32(cycle_speed_name))
		msg->FindInt32(cycle_speed_name, &settings->cycle_speed);
	if (msg->HasInt32(drawing_gap_name))
		msg->FindInt32(drawing_gap_name, &settings->drawing_gap);
	if (msg->HasInt32(drawing_speed_name))
		msg->FindInt32(drawing_speed_name, &settings->drawing_speed);
	if (msg->HasInt32(screen_resolution_name))
		msg->FindInt32(screen_resolution_name,
					   (int32 *)&settings->screen_resolution);
	if (msg->HasBool(compatibility_mode_name))
		msg->FindBool(compatibility_mode_name, &settings->compatibility_mode);
	srand(real_time_clock());
}

// destructor
BrainWashSaver::~BrainWashSaver()
{
	delete settings;
}

// StartSaver
status_t BrainWashSaver::StartSaver(BView* view, bool preview)
{
	if (preview) {
		fDisplay = new ViewDisplay(view);
	} else {
		if (settings->compatibility_mode) {
			fDisplay = new ViewDisplay(view);
		} else {
			// find out desired resolution
			uint32 res = settings->screen_resolution;
			if (res == CURRENT_WORKSPACE_DIMENSIONS)
			{
				// current resolution
				BRect frame = BScreen().Frame();
				res = color_space_for(frame.IntegerWidth() + 1,
									  frame.IntegerHeight() + 1);
			}
			// init the display and start the threads
			fDisplay = new WindowScreenDisplay(res);
		}
	}
	return B_OK;
}

// StopSaver
void BrainWashSaver::StopSaver(void)
{
	// stop the threads
	if (fDrawThread) {
		fDrawThread->Die();
		delete fDrawThread;
		fDrawThread = NULL;
	}
	if (fCycleThread) {
		fCycleThread->Die();
		delete fCycleThread;
		fCycleThread = NULL;
	}
	// close the display
	if (fDisplay) {
		fDisplay->Cleanup();
		fDisplay = NULL;
	}
}

// StartConfig
void BrainWashSaver::StartConfig(BView *configView)
{
	fLooper = new BrainLooper(this);
	fLooper->Run();

	// title
	BTabView* tv_settings = new BTabView(configView->Bounds(), "tab_view");
	tv_settings->SetViewColor(configView->ViewColor());
	configView->AddChild(tv_settings);
	BRect tabframe = tv_settings->Bounds();
	tabframe.right -= 2;
	tabframe.bottom -= tv_settings->TabHeight() + 2;

	BView* page1 = add_tab_page(tv_settings, tabframe, "Credits");
	BView* page2 = add_tab_page(tv_settings, tabframe, "Saver");
	BView* page3 = add_tab_page(tv_settings, tabframe, "Screen");
	tv_settings->SetTabWidth(B_WIDTH_FROM_WIDEST);

	tabframe.InsetBy(5, 5);

	// the first page
	BRect frame = tabframe;
	frame.top += 5;
	frame.left += 7.0;
	// heading
	BFont font(be_bold_font);
	font.SetSize(18);
	font_height fontHeight;
	font.GetHeight(&fontHeight);
	frame.bottom = frame.top + fontHeight.ascent + fontHeight.descent;
	BStringView *strview = 
		new BStringView(frame, B_EMPTY_STRING, cmd_version);
	page1->AddChild(strview);
	strview->SetViewColor(configView->ViewColor());
	strview->SetFont(&font);
	// info text
	frame.top = frame.bottom + 7.0;
	frame.bottom = frame.top + 100.0;
	BRect textArea = frame;
	textArea.OffsetTo(B_ORIGIN);
	BTextView *caption = new BTextView(frame, B_EMPTY_STRING, textArea,
						   			   B_FOLLOW_ALL, B_WILL_DRAW);
	font = *be_plain_font;
	font.SetSize(13);
	caption->SetFontAndColor(&font);
	caption->SetText("© 1998-2003 \n"
					 "Stephan Aßmus, Ingo Weinhold\n"
					 "<stippi@yellowbites.com>\n"
					 "<bonefish@yellowbites.com>\n");
	page1->AddChild(caption);
	caption->SetViewColor(configView->ViewColor());
	caption->SetDrawingMode(B_OP_OVER);
	caption->SetFont(be_plain_font);
	caption->MakeEditable(FALSE);
	caption->MakeSelectable(TRUE);
	caption->SetWordWrap(TRUE);	

	// the second page
	frame = tabframe;
	// the cycle_speed slider
	frame.bottom = frame.top + 40.0;
	page2->AddChild(new SimpleSlider(frame, "cycle speed: ", MIN_CYCLE_SPEED,
		MAX_CYCLE_SPEED, settings->cycle_speed, "slow", "fast",
		&settings->cycle_speed));
	// the drawing_gap slider
	frame.top = frame.bottom + 10;
	frame.bottom = frame.top + 40.0;
	page2->AddChild(new SimpleSlider(frame, "drawing gap (secs): ",
		MIN_DRAWING_GAP, MAX_DRAWING_GAP, settings->drawing_gap, "0", "10",
		&settings->drawing_gap));
	// the drawing_speed slider
	frame.top = frame.bottom + 10;
	frame.bottom = frame.top + 40.0;
	page2->AddChild(new SimpleSlider(frame, "drawing speed: ",
		MIN_DRAWING_SPEED, MAX_DRAWING_SPEED, settings->drawing_speed, "slow",
		"fast", &settings->drawing_speed));

	// the third page
	frame = tabframe;
	// compatibility mode check box
	frame.top += 10;
	frame.bottom = frame.top + 20.0;
	frame.right = frame.left + 200.0;
	fCompatibilityModeBox = new SimpleCheckBox(frame,
		"compatibility mode (CPU hungry!)",
		settings->compatibility_mode, &settings->compatibility_mode,
		new BMessage(MSG_COMPATIBILITY_MODE_CHANGED));
	fCompatibilityModeBox->SetTarget(fLooper);
	page3->AddChild(fCompatibilityModeBox);
	// the screen resolution menu
	frame.top = frame.bottom + 10;
	frame.bottom = frame.top + 20.0;
	frame.right = frame.left + 200.0;
	fScreenResolutionMenu = new ScreenResMenuField(frame, "resolution:",
		settings->screen_resolution, &settings->screen_resolution);
	fScreenResolutionMenu->SetEnabled(!settings->compatibility_mode);
	page3->AddChild(fScreenResolutionMenu);

	// select a page in the tab view
	tv_settings->Select(0);
}

// StopConfig
void BrainWashSaver::StopConfig()
{
	// terminate the looper first to avoid race conditions
	// (CompatibilityModeChanged() is called by the looper thread)
	if (fLooper) {
		if (fLooper->Lock())
			fLooper->Quit();
		fLooper = NULL;
	}
	fScreenResolutionMenu = NULL;
	fCompatibilityModeBox = NULL;
}

// Draw
void BrainWashSaver::Draw(BView *view, int32 frame)
{
	if (frame == 0)
	{
		// if screen_blanker's BDirectWindow is in full screen mode, we cannot
		// open our BWindowScreen, so set it to window mode
		if (BDirectWindow *win = dynamic_cast<BDirectWindow *>(view->Window()))
			win->SetFullScreen(false);
		// init the display and start the threads
		status_t error = fDisplay->Init();
		if (error == B_OK) {
			fCycleThread = new CycleThread(fDisplay);
			fDrawThread = new DrawThread(fDisplay, fCycleThread);
		}
	}
	if (fDisplay)
		fDisplay->Refresh(frame);
}

// SaveState
status_t BrainWashSaver::SaveState(BMessage *state) const
{
	state->AddInt32(cycle_speed_name, settings->cycle_speed);
	state->AddInt32(drawing_gap_name, settings->drawing_gap);
	state->AddInt32(drawing_speed_name, settings->drawing_speed);
	state->AddInt32(screen_resolution_name,
					(int32)settings->screen_resolution);
	state->AddBool(compatibility_mode_name,
				   (int32)settings->compatibility_mode);
	return B_OK;
}

// CompatibilityModeChanged
void
BrainWashSaver::CompatibilityModeChanged()
{
	if (fScreenResolutionMenu && fScreenResolutionMenu->LockLooper()) {
		bool mode = (fCompatibilityModeBox->Value() == B_CONTROL_ON);
		fScreenResolutionMenu->SetEnabled(!mode);
		if (mode)
			fScreenResolutionMenu->SetResolution(CURRENT_WORKSPACE_DIMENSIONS);
		fScreenResolutionMenu->UnlockLooper();
	}
}


// instantiate_screen
BScreenSaver *instantiate_screen_saver(BMessage *msg, image_id id)
{
	return new BrainWashSaver(msg, id);
}

