// ViewDisplay.cpp

#include <string.h>

#include <Autolock.h>
#include <Bitmap.h>
#include <Screen.h>
#include <View.h>

#include "ViewDisplay.h"

// constructor
ViewDisplay::ViewDisplay(BView* view)
	: Display(),
	  fView(view),
	  fFrameBuffer(NULL),
	  fTrueColorBitmap(NULL),
	  fPaletteLock()
{
}

// destructor
ViewDisplay::~ViewDisplay()
{
}

// Init
status_t
ViewDisplay::Init()
{
	// get the view's dimensions
	if (fView == NULL)
		return B_BAD_VALUE;
	if (!fView->LockLooper())
		return B_ERROR;
	BRect bounds = fView->Bounds();
	fView->UnlockLooper();
	int32 width = bounds.IntegerWidth() + 1;
	int32 height = bounds.IntegerHeight() + 1;
	// create the bitmap and the framebuffer
	fTrueColorBitmap = new BBitmap(BRect(0, 0, width - 1, height - 1),
								   B_RGB32);
	uint8* bits = new uint8[width * height];
	fFrameBuffer = new FrameBuffer(bits, width, height, width);
	return B_OK;
}

// Cleanup
void
ViewDisplay::Cleanup()
{
	// delete the frame buffer
	if (fFrameBuffer) {
		uint8* bits = fFrameBuffer->GetBits();
		delete fFrameBuffer;
		fFrameBuffer = NULL;
		delete[] bits;
	}
	// delete the bitmap
	if (fTrueColorBitmap) {
		delete fTrueColorBitmap;
		fTrueColorBitmap = NULL;
	}
}

// GetFrameBuffer
FrameBuffer
ViewDisplay::GetFrameBuffer()
{
	return *fFrameBuffer;
}

// SetPalette
void
ViewDisplay::SetPalette(const rgb_color *colors)
{
	BAutolock _(fPaletteLock);
	memcpy(&fPalette, colors, sizeof(rgb_color[256]));
}

// GetPalette
void
ViewDisplay::GetPalette(rgb_color *colors)
{
	BAutolock _(fPaletteLock);
	memcpy(colors, &fPalette, sizeof(rgb_color[256]));
}

// Refresh
void
ViewDisplay::Refresh(int32 /*frame*/)
{
	if (!fView)
		return;
	// get the palette
	rgb_color palette[256];
	GetPalette(palette);
	// translate the 8 bit frame buffer to the true color bitmap
	int32 width = fFrameBuffer->Width();
	int32 height = fFrameBuffer->Height();
	int32 inBPR = fFrameBuffer->BPR();
	int32 outBPR = fTrueColorBitmap->BytesPerRow();
	uint8* inBits = fFrameBuffer->GetBits();
	uint8* outBits = (uint8*)fTrueColorBitmap->Bits();
	uint8* inLine = inBits;
	uint8* outLine = outBits;
	// B[7:0]  G[7:0]  R[7:0]  -[7:0]
	struct pixel_t {
		uint8	blue;
		uint8	green;
		uint8	red;
		uint8	pad;
	};
	for (int32 y = 0; y < height; y++, inLine += inBPR, outLine += outBPR) {
		uint8* inPixel = inLine;
		pixel_t* outPixel = (pixel_t*)outLine;
		for (int32 x = 0; x < width; x++, inPixel++, outPixel++) {
			rgb_color& color = palette[*inPixel];
			outPixel->red = color.red;
			outPixel->green = color.green;
			outPixel->blue = color.blue;
		}
	}
	// wait for vertical blank (to avoid flickering) and draw the bitmap
	BScreen().WaitForRetrace();
	fView->DrawBitmap(fTrueColorBitmap);
}

